import os from "os";
import fs from "fs";
import path from "path";
import { getConfig } from "./config";


const URLS = {
	GITHUB: "https://github.com/kangkang520/node-addons/releases/download/ai{{version}}/{{filename}}",
	XXXXX: "http://git.urnas.cn:5200/yizhi-js-lib/ai-box/releases/download/{{version}}/{{filename}}",
}

function releaseVersion() { return require("../../package.json").releaseVersion }

function getURL(backend: "ort" | "mnn", template: string) {

	const URL_DICT: Record<string, Record<string, string>> = {
		"win32": {
			"x64": `${backend}_windows_x64.node`
		},
		"linux": {
			"x64": `${backend}_linux_x64.node`,
			"arm64": `${backend}_linux_arm64.node`,
		},
		"darwin": {
			"arm64": `${backend}_macos_arm64.node`,
		}
	}

	const archConfig = URL_DICT[os.platform()];
	if (!archConfig) throw new Error(`Unsupported platform: ${os.platform()}`);
	const downloadName = archConfig[os.arch()];
	if (!downloadName) throw new Error(`Unsupported arch: ${os.arch()}`);

	return template.replaceAll("{{version}}", releaseVersion()).replaceAll("{{filename}}", downloadName);
}

async function getStream(backend: "mnn" | "ort") {
	for (const [name, url] of Object.entries(URLS)) {
		try {
			return await fetch(getURL(backend, url)).then(res => {
				if (res.status != 200) throw new Error("Failed to download addon.");
				return res.blob().then(b => b.stream());
			})
		} catch (e) { }
	}
	throw new Error("Failed to download addon.");
}

export async function downloadBackend(backend: "ort" | "mnn", savename?: string) {
	const backendConfigNameDict = { ort: "ORT_ADDON_FILE" as const, mnn: "MNN_ADDON_FILE" as const };
	const defaultAddon = path.resolve(process.cwd(), getConfig(backendConfigNameDict[backend]));

	const saveName = savename ? path.resolve(path.dirname(defaultAddon), savename) : defaultAddon;
	if (fs.existsSync(saveName)) {
		try {
			const addon = require(saveName);
			if (addon.__release__ === releaseVersion()) return saveName;
			//清除缓存
			delete require.cache[saveName];
		} catch (err) { }
	}

	await fs.promises.mkdir(path.dirname(saveName), { recursive: true });

	const stream = await getStream(backend);
	const cacheFile = await new Promise<string>((resolve, reject) => {
		const cacheFile = path.join(os.tmpdir(), Date.now() + ".cv.node");
		let fsStream!: ReturnType<typeof fs.createWriteStream>;
		stream.pipeTo(new WritableStream({
			start(controller) {
				fsStream = fs.createWriteStream(cacheFile);
			},
			async write(chunk, controller) {
				await new Promise<void>((resolve, reject) => fsStream.write(chunk, err => err ? reject(err) : resolve()));
			},
			close() {
				fsStream.end();
				resolve(cacheFile);
			},
			abort() { }
		})).catch(reject);
	});
	if (fs.existsSync(saveName)) await fs.promises.rm(saveName, { recursive: true, force: true });
	await fs.promises.cp(cacheFile, saveName);
	await fs.promises.rm(cacheFile);
	return saveName;
}